import argparse
import unittest
from unittest.mock import Mock, patch

import tufup
import tufup.repo.cli
import tufup.utils
from tests import TempDirTestCase


class ParserTests(unittest.TestCase):
    def test_get_parser(self):
        parser = tufup.repo.cli.get_parser()
        for cmd in [
            'init',
            'init --debug',
            'targets add 1.0 c:\\my_bundle_dir c:\\private_keys',
            'targets -d add 1.0 c:\\my_bundle_dir c:\\private_keys',
            'targets -d add -s 1.0 c:\\my_bundle_dir c:\\private_keys',
            'targets remove-latest c:\\private_keys',
            'keys my-key-name -c -e',
            'keys my-key-name add root c:\\private_keys d:\\more_private_keys',
            'keys my-key-name -c -e add root c:\\private_keys',
            'keys my-key-name replace old-key-name c:\\private_keys',
            'keys my-key-name -c -e replace old-key-name c:\\private_keys',
            'sign root c:\\private_keys d:\\other_private_keys',
            'sign root c:\\private_keys -e',
            'sign root c:\\private_keys -e 100',
        ]:
            with self.subTest(msg=cmd):
                args = cmd.split()
                options = parser.parse_args(args)
                expected_func_name = '_cmd_' + args[0]
                self.assertEqual(expected_func_name, options.func.__name__)


class CommandTests(TempDirTestCase):
    def setUp(self) -> None:
        super().setUp()
        role_names = ['root', 'targets', 'snapshot', 'timestamp']
        self.config = dict(
            app_name='my-app',
            repo_dir=self.temp_dir_path / 'repo',
            keys_dir=self.temp_dir_path / 'keys',
            key_map={name: [name] for name in role_names},
            encrypted_keys=[],
            expiration_days={name: 1 for name in role_names},
        )
        mock_return_config = Mock(return_value=self.config)
        mock_keys = Mock()
        mock_keys.create_key_pair = Mock()
        mock_repo = Mock(keys=mock_keys, **self.config)
        mock_repo.save_config = Mock()
        mock_repo.initialize = Mock()
        mock_repo.add_bundle = Mock()
        mock_repo.remove_latest_bundle = Mock()
        mock_repo.add_key = Mock()
        mock_repo.replace_key = Mock()
        mock_repo.refresh_expiration_date = Mock()
        mock_repo.threshold_sign = Mock()
        mock_repo.publish_changes = Mock()
        MockRepository = Mock(return_value=mock_repo)  # noqa
        MockRepository.load_config = mock_return_config
        MockRepository.from_config = Mock(return_value=mock_repo)
        self.mock_repo = mock_repo
        self.mock_repo_class = MockRepository

    def test__cmd_init(self):
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            with patch('tufup.repo.cli.input_bool', Mock(return_value=True)):
                with patch(
                        'tufup.repo.cli._get_config_from_user',
                        self.mock_repo_class.load_config,
                ):
                    tufup.repo.cli._cmd_init(options=argparse.Namespace())
        self.mock_repo.initialize.assert_called()

    def test__cmd_keys_create(self):
        options = argparse.Namespace(
            new_key_name='test', encrypted=True, create=True
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_keys(options=options)
        self.mock_repo.keys.create_key_pair.assert_called()

    def test__cmd_keys_create_and_add_key(self):
        options = argparse.Namespace(
            create=True,
            encrypted=True,
            key_dirs=['c:\\my_private_keys'],
            new_key_name='test',
            role_name='root',
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_keys(options=options)
        self.mock_repo.keys.create_key_pair.assert_called()
        self.mock_repo.add_key.assert_called()
        self.mock_repo.publish_changes.assert_called()

    def test__cmd_keys_replace_key(self):
        options = argparse.Namespace(
            create=True,
            encrypted=False,
            key_dirs=['c:\\my_private_keys'],
            new_key_name='some new key to be created',
            old_key_name='some old key name',
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_keys(options=options)
        self.mock_repo.replace_key.assert_called()
        self.mock_repo.keys.create_key_pair.assert_called()
        self.mock_repo.publish_changes.assert_called()

    def test__cmd_targets_add(self):
        version = '1.0'
        bundle_dir = 'dummy'
        key_dirs = ['c:\\my_private_keys']
        skip_patch = True
        options = argparse.Namespace(
            app_version=version,
            bundle_dir=bundle_dir,
            key_dirs=key_dirs,
            skip_patch=skip_patch,
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_targets(options=options)
        self.mock_repo.add_bundle.assert_called_with(
            new_version=version, new_bundle_dir=bundle_dir, skip_patch=skip_patch,
        )
        self.mock_repo.publish_changes.assert_called_with(
            private_key_dirs=key_dirs
        )

    def test__cmd_targets_remove_latest(self):
        key_dirs = ['c:\\my_private_keys']
        options = argparse.Namespace(key_dirs=key_dirs)
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_targets(options=options)
        self.mock_repo.remove_latest_bundle.assert_called()

    def test__cmd_sign_threshold(self):
        role_name = 'root'
        key_dirs = ['c:\\my_private_keys']
        options = argparse.Namespace(
            role_name=role_name, key_dirs=key_dirs, expiration_days=None
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_sign(options=options)
        self.mock_repo.threshold_sign.assert_called_with(
            role_name=role_name, private_key_dirs=key_dirs
        )

    def test__cmd_sign_expired(self):
        role_name = 'root'
        key_dirs = ['c:\\my_private_keys']
        options = argparse.Namespace(
            role_name=role_name,
            key_dirs=key_dirs,
            expiration_days='default',  # i.e. specify -e without a value
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_sign(options=options)
        self.mock_repo.refresh_expiration_date.assert_called_with(
            role_name=role_name, days=self.config['expiration_days'][role_name]
        )
        self.mock_repo.publish_changes.assert_called_with(
            private_key_dirs=key_dirs
        )

    def test__get_config_from_user_no_kwargs(self):
        default = ''
        yes = 'y'
        no = 'n'
        user_inputs = iter(
            [
                'my-app',  # app name
                'my_app.__version__',  # app version
                'repo/dir',  # repo dir
                'keys/dir',  # keys dir
                yes,  # keep default root key name
                yes,  # add another root key name
                'root-2',  # key name
                no,  # add another root key name
                default,  # encrypt 'root' key
                yes,  # encrypt 'root-2' key
                '365',  # expiration days
                '2',  # signature threshold
                default,  # keep default targets key name
                default,  # add another targets key name
                default,  # encrypt 'targets' key
                '30',  # expiration days
                '1',  # signature threshold
                default,  # keep default snapshot key name
                default,  # add another snapshot key name
                default,  # encrypt 'snapshot' key
                '7',  # expiration days
                '1',  # signature threshold
                default,  # keep default timestamp key name
                default,  # add another timestamp key name
                default,  # encrypt 'timestamp' key
                '1',  # expiration days
                '1',  # signature threshold
            ]
        )
        with patch('builtins.input', lambda *_, **__: next(user_inputs)):
            config_kwargs = tufup.repo.cli._get_config_from_user()
        self.assertTrue(config_kwargs)

    def test__get_config_from_user_with_kwargs(self):
        role_names = ['root', 'targets', 'snapshot', 'timestamp']
        original_kwargs = dict(
            app_name='my-app',
            app_version_attr='my_app.__version__',
            repo_dir='repo/dir',
            keys_dir='keys/dir',
            key_map={name: [name] for name in role_names},
            encrypted_keys=['root'],
            expiration_days={name: 1 for name in role_names},
            thresholds={name: 1 for name in role_names},
        )
        default = ''
        with patch('builtins.input', Mock(return_value=default)):
            config_kwargs = tufup.repo.cli._get_config_from_user(
                **original_kwargs
            )
        self.assertEqual(config_kwargs, original_kwargs)



from unittest import TestCase

from packaging.version import Version
import tuf


class PackageTests(TestCase):
    def test_tuf_version(self):
        # quick & dirty regression test for issue 44
        self.assertGreaterEqual(Version(tuf.__version__), Version('2'))



import pathlib
import unittest
from unittest.mock import Mock, patch

import tufup.utils
from tests import TempDirTestCase


class RemovePathTests(TempDirTestCase):
    def test_remove_path(self):
        for arg_type in [str, pathlib.Path]:
            # create a directory and subdirectory with dummy files
            dir_path = self.temp_dir_path / 'dir'
            subdir_path = dir_path / 'subdir'
            subdir_path.mkdir(parents=True)
            for file in [dir_path / 'file.in.dir', subdir_path / 'file.in.subdir']:
                file.touch()
            # just to be sure
            self.assertEqual(2, len(list(dir_path.iterdir())))
            self.assertEqual(1, len(list(subdir_path.iterdir())))
            # test
            with self.subTest(msg=arg_type):
                self.assertTrue(
                    tufup.utils.remove_path(path=arg_type(dir_path))
                )
                self.assertFalse(dir_path.exists())


class InputTests(unittest.TestCase):
    def test_input_bool(self):
        inputs = [('', None), ('y', True), ('n', False), ('anything', False)]
        for default in [True, False]:
            for user_input, expected in inputs:
                if expected is None:
                    expected = default
                with patch('builtins.input', Mock(return_value=user_input)):
                    self.assertEqual(
                        expected,
                        tufup.utils.input_bool(prompt='', default=default),
                    )

    def test_input_list(self):
        default = ['existing item']
        item_default = 'default item'
        new_item = 'new item'
        bool_inputs = iter([True, True, True, False])
        text_inputs = iter(['', new_item])
        # we use iterators to simulate sequences of user inputs
        with patch.object(
                tufup.utils, 'input_bool', lambda *_, **__: next(bool_inputs)
        ):
            with patch.object(
                    tufup.utils,
                    'input_text',
                    lambda *_, **__: next(text_inputs) or item_default,
            ):
                expected = default + [item_default, new_item]
                self.assertEqual(
                    expected,
                    tufup.utils.input_list(
                        prompt='', default=default, item_default=item_default
                    )
                )

    def test_input_numeric(self):
        default = 1
        answer = 0
        user_inputs = iter(['not numeric', str(answer)])
        # we use an iterator to simulate a sequence of user inputs,
        # and return '' instead of raising StopIteration
        with patch('builtins.input', lambda *_: next(user_inputs, '')):
            self.assertEqual(
                answer, tufup.utils.input_numeric(prompt='', default=default)
            )
            # iterator exhausted, so next user input is ''
            self.assertEqual(
                default, tufup.utils.input_numeric(prompt='', default=default)
            )

    def test_input_text(self):
        answer = 'something'
        user_inputs = iter(['', answer])
        with patch('builtins.input', lambda *_: next(user_inputs, '')):
            # this should iterate until we get a non-empty answer
            self.assertEqual(
                answer, tufup.utils.input_text(prompt='', default='')
            )
            # iterator exhausted, so next user input is ''
            self.assertEqual(
                answer, tufup.utils.input_text(prompt='', default=answer)
            )

    def test_input_text_optional(self):
        with patch('builtins.input', Mock(return_value='')):
            self.assertIsNone(
                tufup.utils.input_text(
                    prompt='', default=None, optional=True
                )
            )



import pathlib

import bsdiff4
from packaging.version import Version

from tests import TempDirTestCase
from tufup.common import Patcher, TargetMeta


class TestTargetMeta(TempDirTestCase):
    def test_eq_ne(self):
        target_meta = TargetMeta()
        self.assertEqual(target_meta, TargetMeta())
        self.assertNotEqual(target_meta, TargetMeta(target_path='something'))
        # two ways to initialize a TargetMeta instance:
        self.assertEqual(
            TargetMeta(target_path='my-app-1.2.3a0.patch'),
            TargetMeta(name='my-app', version='1.2.3a0', is_archive=False),
        )

    def test_repr(self):
        # https://docs.python.org/3/reference/datamodel.html#object.__repr__
        target_meta = TargetMeta(target_path='some.file')
        # the input to eval is known and trusted, so that should be safe, right?
        self.assertEqual(eval(repr(target_meta)), target_meta)

    def test_str(self):
        # see issue #3
        for target_path in ['str_path', pathlib.Path('pathlib_path')]:
            with self.subTest(msg=target_path):
                self.assertIsInstance(
                    TargetMeta(target_path=target_path).__str__(), str
                )

    def test_hashable(self):
        obj = TargetMeta()
        self.assertEqual(obj.__hash__(), hash(tuple(vars(obj).items())))
        # we can use the obj as a set member or as dict key
        self.assertEqual({obj, obj}, {obj})

    def test_sortable(self):
        version_1 = TargetMeta(target_path='my-app-win-1.0.tar.gz')
        version_2 = TargetMeta(target_path='my-app-win-2.0.tar.gz')
        info_list = [version_2, version_1]
        self.assertEqual([version_1, version_2], sorted(info_list))

    def test_filename(self):
        target_path = 'url/path/somefile'
        target_meta = TargetMeta(target_path=target_path)
        self.assertEqual('somefile', target_meta.filename)

    def test_name(self):
        target_meta = TargetMeta(target_path='url/path/my-app-1.0.tar.gz')
        self.assertEqual('my-app', target_meta.name)

    def test_version(self):
        for version_str in ['1.2.3a4', '']:
            expected = Version(version_str) if version_str else None
            with self.subTest(msg=version_str):
                target_meta = TargetMeta(target_path=f'x-{version_str}.tar.gz')
                self.assertEqual(expected, target_meta.version)

    def test_suffix(self):
        valid_suffixes = ['.tar.gz', '.patch']
        for suffix in ['', '.zip', '.tar.gz', '.patch']:
            with self.subTest(msg=suffix):
                target_meta = TargetMeta(target_path=f'my-app-1.2.3a4{suffix}')
                if suffix in valid_suffixes:
                    self.assertEqual(suffix, target_meta.suffix)
                else:
                    self.assertIsNone(target_meta.suffix)

    def test_is_archive(self):
        self.assertTrue(TargetMeta(target_path='my-app-1.0.tar.gz').is_archive)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.patch').is_archive)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.zip').is_archive)

    def test_is_patch(self):
        self.assertTrue(TargetMeta(target_path='my-app-1.0.patch').is_patch)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.tar.gz').is_patch)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.zip').is_patch)

    def test_is_other(self):
        self.assertTrue(TargetMeta(target_path='my-app-1.0.zip').is_other)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.patch').is_other)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.tar.gz').is_other)

    def test_parse_filename(self):
        cases = [
            # PEP440 versions are allowed (we do not parse them here...)
            ('app-1.patch', ('app', '1', '.patch')),
            ('app-1.tar.gz', ('app', '1', '.tar.gz')),
            ('app-1.2.tar.gz', ('app', '1.2', '.tar.gz')),
            ('app-1.2.3.tar.gz', ('app', '1.2.3', '.tar.gz')),
            ('app-1a.tar.gz', ('app', '1a', '.tar.gz')),
            ('app-1b.tar.gz', ('app', '1b', '.tar.gz')),
            ('app-1rc.tar.gz', ('app', '1rc', '.tar.gz')),
            ('app-1rc0.tar.gz', ('app', '1rc0', '.tar.gz')),
            ('app-2022.0.tar.gz', ('app', '2022.0', '.tar.gz')),
            # we don't impose a specific version format at this point (that
            # is deferred to packaging.Version)
            ('app-invalidversion.tar.gz', ('app', 'invalidversion', '.tar.gz')),
            # underscores, dashes, capitals, etc.
            ('app-name---1.tar.gz', ('app-name--', '1', '.tar.gz')),
            ('CAPS-1.tar.gz', ('CAPS', '1', '.tar.gz')),
            ('un_der_scores-1.0.tar.gz', ('un_der_scores', '1.0', '.tar.gz')),
            # invalid filenames
            ('sp ac es-1.zip', ()),
            ('app-1.gz', ()),
            ('app-1.xz', ()),
            ('anything', ()),
        ]
        for filename, expected in cases:
            match_dict = TargetMeta.parse_filename(filename=filename)
            with self.subTest(msg=filename):
                self.assertEqual(expected, tuple(match_dict.values()))

    def test_compose_filename(self):
        filename = TargetMeta.compose_filename(
            name='app', version='1.0', is_archive=True
        )
        self.assertEqual('app-1.0.tar.gz', filename)


class PatcherTests(TempDirTestCase):
    def setUp(self) -> None:
        super().setUp()
        # dummy paths
        self.old_archive_path = self.temp_dir_path / 'my_app-1.0.tar.gz'
        self.new_archive_path = self.temp_dir_path / 'my_app-2.0.tar.gz'
        self.new_patch_path = self.temp_dir_path / 'my_app-2.0.patch'
        # write dummy archive data to files
        self.old_archive_path.write_bytes(b'old archive data')
        self.new_archive_data = b'new archive data'
        self.new_archive_path.write_bytes(self.new_archive_data)
        # create patch file (see Patcher.create_patch)
        bsdiff4.file_diff(
            src_path=self.old_archive_path,
            dst_path=self.new_archive_path,
            patch_path=self.new_patch_path,
        )
        self.new_patch_data = self.new_patch_path.read_bytes()

    def test_create_patch(self):
        # remove existing patch file, just to be sure
        self.new_patch_path.unlink()
        # test
        new_patch_path = Patcher.create_patch(
            src_path=self.old_archive_path, dst_path=self.new_archive_path
        )
        self.assertTrue(new_patch_path.exists())
        self.assertEqual(self.new_patch_data, new_patch_path.read_bytes())

    def test_apply_patch(self):
        # remove existing "new archive" file, just to be sure
        self.new_archive_path.unlink()
        # test
        new_archive_path = Patcher.apply_patch(
            src_path=self.old_archive_path, patch_path=self.new_patch_path
        )
        self.assertTrue(new_archive_path.exists())
        self.assertEqual(self.new_archive_data, new_archive_path.read_bytes())



import copy
from datetime import date, datetime, timedelta
import json
import logging
import pathlib
from tempfile import TemporaryDirectory
from time import sleep
from unittest.mock import Mock, patch

from securesystemslib.interface import (
    generate_and_write_unencrypted_ed25519_keypair,
    generate_and_write_ed25519_keypair,
    import_ed25519_publickey_from_file,
)
from tuf.api.metadata import (
    Metadata,
    Role,
    Root,
    Targets,
    Snapshot,
    Timestamp,
    TargetFile,
    TOP_LEVEL_ROLE_NAMES
)

from tests import TempDirTestCase, TEST_REPO_DIR
from tufup.common import TargetMeta
import tufup.repo  # for patching
from tufup.repo import (
    Base, in_, Keys, make_gztar_archive, Repository, Roles, SUFFIX_PUB, SUFFIX_PATCH
)


mock_input = Mock(return_value='')

DUMMY_SSL_KEY = {
    'keytype': 'ed25519',
    'scheme': 'ed25519',
    'keyid': '22f7c6046e29cfb0205a1c07941a5a57da39a6859b844f8c347f622a57ff82c8',
    'keyid_hash_algorithms': ['sha256', 'sha512'],
    'keyval': {'public': '93032b5804ba40a725145171193782bdfa30038584715546aea3228ea8018e46'},
}
DUMMY_ROOT = Root(
    version=1,
    spec_version='1.0',
    expires=datetime.now() + timedelta(days=1),
    keys=dict(),
    roles={
        role_name: Role(keyids=[], threshold=1)
        for role_name in TOP_LEVEL_ROLE_NAMES
    },
    consistent_snapshot=False,
)
DUMMY_EXPIRATION_DAYS = dict(root=1000, targets=100, snapshot=10, timestamp=1)
DUMMY_PRIVATE_KEY_PATHS = dict(
    (role_name, [pathlib.Path('dummy', role_name)])
    for role_name in TOP_LEVEL_ROLE_NAMES
)
DUMMY_KEY_MAP = dict(
    root=['root_one', 'root_two'],
    targets=['targets'],
    snapshot=['snapshot'],
    timestamp=['timestamp'],
)
DUMMY_THRESHOLDS = dict(root=4, targets=3, snapshot=2, timestamp=1)


class ModuleTests(TempDirTestCase):
    def test_in_(self):
        self.assertIsInstance(in_(days=1), datetime)

    def test_make_gztar_archive(self):
        app_name = 'test'
        version = '1.2.3'
        # prepare
        sub_dir = self.temp_dir_path / 'sub'
        sub_file = sub_dir / 'sub.txt'
        root_file = self.temp_dir_path / 'root.txt'
        sub_dir.mkdir()
        sub_file.touch()
        root_file.touch()
        # test
        mock_input_no = Mock(return_value='n')
        for exists in [False, True]:
            with patch('builtins.input', mock_input_no):
                archive = make_gztar_archive(
                    src_dir=self.temp_dir_path,
                    dst_dir=self.temp_dir_path,
                    app_name=app_name,
                    version=version,
                    base_dir='.',  # this kwarg is allowed
                    root_dir='some path',  # this kwarg is removed
                )
            self.assertIsInstance(archive, TargetMeta)
            self.assertEqual(exists, mock_input_no.called)
            self.assertTrue(archive.path.exists())
            self.assertTrue(app_name in str(archive.path))
            self.assertTrue(version in str(archive.path))


class BaseTests(TempDirTestCase):
    def test_init(self):
        with patch('builtins.input', mock_input):
            # dir exists
            base = Base(dir_path=self.temp_dir_path)
            self.assertTrue(base.dir_path.exists())
            # dir does not exist yet
            base = Base(dir_path=self.temp_dir_path / 'new')
            self.assertTrue(base.dir_path.exists())


class KeysTests(TempDirTestCase):
    def test_init_no_key_files(self):
        # no public key files exist yet
        keys = Keys(dir_path=self.temp_dir_path)
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertFalse(getattr(keys, role_name))

    def test_init_and_import_all_public_keys(self):
        # create some key files
        for role_name in TOP_LEVEL_ROLE_NAMES:
            private_key_filename = Keys.filename_pattern.format(key_name=role_name)
            file_path = self.temp_dir_path / private_key_filename
            generate_and_write_unencrypted_ed25519_keypair(filepath=str(file_path))
        # test
        keys = Keys(dir_path=self.temp_dir_path)
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertIsInstance(getattr(keys, role_name)[0], dict)

    def test_init_and_import_all_public_keys_with_key_map(self):
        # create a single key-pair
        key_name = 'single'
        private_key_filename = Keys.filename_pattern.format(key_name=key_name)
        file_path = self.temp_dir_path / private_key_filename
        generate_and_write_unencrypted_ed25519_keypair(filepath=str(file_path))
        key_map = {role_name: [key_name] for role_name in TOP_LEVEL_ROLE_NAMES}
        # test
        keys = Keys(dir_path=self.temp_dir_path, key_map=key_map)
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertIsInstance(getattr(keys, role_name)[0], dict)
            # all keys should be equal
            self.assertEqual(keys.root, getattr(keys, role_name))

    def test_import_all_public_keys(self):
        # note: import_all_public_keys is also (implicitly) tested via __init__
        keys = Keys(dir_path=self.temp_dir_path)
        # create some key files
        for role_name in TOP_LEVEL_ROLE_NAMES:
            private_key_filename = Keys.filename_pattern.format(
                key_name=role_name)
            file_path = self.temp_dir_path / private_key_filename
            generate_and_write_unencrypted_ed25519_keypair(
                filepath=str(file_path))
        # test
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertFalse(getattr(keys, role_name))
        keys.import_all_public_keys()
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertIsInstance(getattr(keys, role_name)[0], dict)

    def test_import_public_key(self):
        # create dummy key with name differing from role name
        key_name = 'test'
        role_name = 'targets'
        private_key_filename = Keys.filename_pattern.format(key_name=key_name)
        file_path = self.temp_dir_path / private_key_filename
        generate_and_write_unencrypted_ed25519_keypair(filepath=str(file_path))
        # test
        keys = Keys(dir_path=self.temp_dir_path)
        keys.import_public_key(role_name=role_name, key_name=key_name)
        self.assertTrue(keys.targets)

    def test_create(self):
        with patch('getpass.getpass', mock_input):
            keys = Keys(dir_path=self.temp_dir_path, encrypted=['root'])
            keys.create()
            # key pair files should now exist
            filenames = [item.name for item in keys.dir_path.iterdir()]
            for role_name in TOP_LEVEL_ROLE_NAMES:
                private_key_filename = Keys.filename_pattern.format(key_name=role_name)
                public_key_filename = private_key_filename + SUFFIX_PUB
                self.assertIn(private_key_filename, filenames)
                self.assertIn(public_key_filename, filenames)
            # and the public keys should have been imported
            self.assertTrue(all(getattr(keys, n) for n in TOP_LEVEL_ROLE_NAMES))

    def test_create_with_key_map(self):
        # prepare
        keys = Keys(dir_path=self.temp_dir_path, key_map=DUMMY_KEY_MAP)
        expected_key_names = [
            key_name
            for key_names in DUMMY_KEY_MAP.values()
            for key_name in key_names
        ]
        # test
        keys.create()
        # we should now have five key-pairs (one for each, but two for root)
        filenames = [item.name for item in keys.dir_path.iterdir()]
        self.assertEqual(2*len(expected_key_names), len(filenames))
        for key_name in expected_key_names:
            self.assertIn(key_name, filenames)
            self.assertIn(key_name + SUFFIX_PUB, filenames)
        # and the public keys should have been imported
        for role_name, key_names in DUMMY_KEY_MAP.items():
            self.assertEqual(len(key_names), len(getattr(keys, role_name)))

    def test_create_key_pair(self):
        public_key_path = Keys.create_key_pair(
            private_key_path=self.temp_dir_path / 'key_name', encrypted=False
        )
        self.assertTrue(public_key_path.exists())

    def test_create_key_pair_do_not_overwrite(self):
        # create dummy key pair
        key_name = 'dummy'
        private_key_filename = Keys.filename_pattern.format(key_name=key_name)
        private_key_path = self.temp_dir_path / private_key_filename
        generate_and_write_unencrypted_ed25519_keypair(
            filepath=str(private_key_path)
        )
        original_private_key = private_key_path.read_bytes()
        with patch('builtins.input', Mock(return_value='n')):
            Keys.create_key_pair(
                private_key_path=private_key_path, encrypted=False
            )
        self.assertEqual(original_private_key, private_key_path.read_bytes())

    def test_public(self):
        keys = Keys(dir_path=self.temp_dir_path)
        # test empty
        self.assertFalse(keys.public())
        # set a dummy key value
        keys.root = [DUMMY_SSL_KEY]
        # test
        self.assertIn(DUMMY_SSL_KEY['keyid'], keys.public().keys())

    def test_roles(self):
        keys = Keys(dir_path=self.temp_dir_path)
        # test empty
        self.assertSetEqual(set(TOP_LEVEL_ROLE_NAMES), set(keys.roles().keys()))
        # set a dummy key value
        keys.root = [DUMMY_SSL_KEY]
        # test
        self.assertIn('root', keys.roles().keys())

    def test_roles_thresholds(self):
        # prepare
        keys = Keys(dir_path=self.temp_dir_path, thresholds=DUMMY_THRESHOLDS)
        for role_name in TOP_LEVEL_ROLE_NAMES:
            setattr(keys, role_name, [DUMMY_SSL_KEY])
        # test
        roles = keys.roles()
        for key, value in roles.items():
            self.assertEqual(DUMMY_THRESHOLDS[key], value.threshold)

    def test_find_private_key(self):
        # create dummy private key files in separate folders
        key_names = [
            ('online', [Snapshot.type, Timestamp.type]),
            ('offline', [Root.type, Targets.type]),
        ]
        key_dirs = []
        for dir_name, role_names  in key_names:
            dir_path = self.temp_dir_path / dir_name
            dir_path.mkdir()
            key_dirs.append(dir_path)
            for role_name in role_names:
                filename = Keys.filename_pattern.format(key_name=role_name)
                (dir_path / filename).touch()
        # test
        for role_name in TOP_LEVEL_ROLE_NAMES:
            key_path = Keys.find_private_key(key_name=role_name, key_dirs=key_dirs)
            self.assertIn(role_name, str(key_path))
            self.assertTrue(key_path.exists())


class RolesTests(TempDirTestCase):
    def test_init(self):
        self.assertTrue(Roles(dir_path=self.temp_dir_path))

    def test_init_import_roles(self):
        def mock_from_file(filename, *args, **kwargs):
            file_path = pathlib.Path(filename)
            assert file_path.exists()
            return file_path.stem

        # create dummy metadata files
        for role_name in TOP_LEVEL_ROLE_NAMES:
            filenames = [f'{role_name}.json']
            if role_name == 'root':
                filenames.extend(f'{v}.{role_name}.json' for v in [1, 2, 3])
            for filename in filenames:
                (self.temp_dir_path / filename).touch()
        # test
        with patch.object(tufup.repo.Metadata, 'from_file', mock_from_file):
            roles = Roles(dir_path=self.temp_dir_path)
            for role_name in TOP_LEVEL_ROLE_NAMES:
                self.assertEqual(role_name, getattr(roles, role_name))

    def test_initialize_empty(self):
        # prepare
        mock_keys = Mock()
        mock_keys.public = Mock()
        mock_keys.roles = Mock(
            return_value={name: None for name in TOP_LEVEL_ROLE_NAMES}
        )
        roles = Roles(dir_path=self.temp_dir_path)
        # test
        roles.initialize(keys=mock_keys)
        self.assertTrue(
            all(
                isinstance(getattr(roles, n), Metadata)
                for n in TOP_LEVEL_ROLE_NAMES
            )
        )
        # files do not exist yet, because the roles still need to be populated
        self.assertFalse(any(roles.dir_path.iterdir()))

    def test_initialize_existing_root(self):
        # prepare
        mock_keys = Mock()
        mock_keys.public = Mock()
        mock_keys.roles = Mock(return_value={n: None for n in TOP_LEVEL_ROLE_NAMES})
        roles = Roles(dir_path=self.temp_dir_path)
        # set existing root role
        mock_root_role = Mock()
        roles.root = mock_root_role
        # test
        roles.initialize(keys=mock_keys)
        # ensure the existing role has not been replaced
        self.assertEqual(mock_root_role, roles.root)

    def test_add_or_update_target(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        roles.targets = Mock(signed=Mock(targets=dict()))
        # test
        filename = 'my_app.tar.gz'
        local_target_path = self.temp_dir_path / filename
        # path must exist
        with self.assertRaises(FileNotFoundError):
            roles.add_or_update_target(local_path=local_target_path)
        local_target_path.write_bytes(b'some bytes')
        # test
        for segments, expected_url_path in [
            (None, filename), ([], filename), (['a', 'b'], 'a/b/' + filename)
        ]:
            roles.add_or_update_target(
                local_path=local_target_path, url_path_segments=segments
            )
            with self.subTest(msg=segments):
                self.assertIsInstance(
                    roles.targets.signed.targets[expected_url_path], TargetFile
                )

    def test_remove_target(self):
        # prepare
        filename = 'my_app-1.0.tar.gz'
        dirname = 'subdir'
        url_path = f'{dirname}/{filename}'
        subdir = self.temp_dir_path / dirname
        subdir.mkdir()
        local_target_path = subdir / filename
        local_target_path.touch()
        roles = Roles(dir_path=self.temp_dir_path)
        roles.targets = Mock(signed=Mock(targets={url_path: Mock()}))
        original_targets_signed = copy.deepcopy(roles.targets.signed)
        # test
        self.assertTrue(local_target_path.exists())
        roles.remove_target(local_path=local_target_path)
        self.assertNotIn(filename, roles.targets.signed.targets)
        self.assertFalse(local_target_path.exists())
        self.assertNotEqual(original_targets_signed, roles.targets.signed)
        # cannot remove non-existent target
        self.assertFalse(roles.remove_target(local_path=local_target_path))

    def test_add_public_key(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        roles.root = Mock(signed=Mock(roles=dict(), add_key=Mock()))
        public_key_path = self.temp_dir_path / 'targets_key.pub'
        public_key_path.write_text(json.dumps(DUMMY_SSL_KEY))
        # test
        role_name = 'targets'
        roles.add_public_key(role_name=role_name, public_key_path=public_key_path)
        self.assertTrue(roles.root.signed.add_key.called)

    def test_set_signature_threshold(self):
        # prepare
        role_name = 'targets'
        threshold = 2
        roles = Roles(dir_path=self.temp_dir_path)
        roles.root = Mock(signed=Mock(roles={role_name: Mock(threshold=1)}))
        # test
        roles.set_signature_threshold(role_name=role_name, threshold=threshold)
        self.assertEqual(threshold, roles.root.signed.roles[role_name].threshold)

    def test_sign_role(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        roles.root = Metadata(signed=DUMMY_ROOT, signatures=dict())
        role_name = 'root'
        signature_count = 2
        password = 'mock-password'
        with patch('getpass.getpass', Mock(return_value=password)):
            for index in range(signature_count):
                private_key_path = self.temp_dir_path / f'{index}.{role_name}'
                # create key pair
                if index == 0:
                    generate_and_write_unencrypted_ed25519_keypair(
                        filepath=str(private_key_path)
                    )
                else:
                    generate_and_write_ed25519_keypair(
                        password=password, filepath=str(private_key_path)
                    )
                # test
                roles.sign_role(
                    role_name=role_name, private_key_path=private_key_path
                )
        self.assertEqual(signature_count, len(roles.root.signatures))

    def test_file_path(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        # test
        self.assertEqual(
            self.temp_dir_path / '1.root.json',
            roles.file_path(role_name='root', version=1),
        )
        self.assertEqual(
            self.temp_dir_path / 'root.json',
            roles.file_path(role_name='root'),
        )
        self.assertEqual(
            self.temp_dir_path / 'targets.json',
            roles.file_path(role_name='targets', version=1),
        )

    def test_file_exists(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        (self.temp_dir_path / '1.root.json').touch()
        # test
        self.assertTrue(roles.file_exists(role_name='root'))
        self.assertFalse(roles.file_exists(role_name='targets'))

    def test_persist_role(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        roles.root = Metadata(signed=DUMMY_ROOT, signatures=dict())
        expected_filenames = [f'{DUMMY_ROOT.version}.root.json', 'root.json']
        # test
        roles.persist_role(role_name='root')
        for filename in expected_filenames:
            with self.subTest(msg=filename):
                self.assertTrue((self.temp_dir_path / filename).exists())

    def test_get_latest_archive(self):
        roles = Roles(dir_path=TEST_REPO_DIR / 'metadata')
        expected_filename = 'example_app-4.0a0.tar.gz'
        latest_archive = roles.get_latest_archive()
        self.assertEqual(expected_filename, latest_archive.path.name)

    def test_get_latest_archive_empty(self):
        roles = Roles(dir_path=self.temp_dir_path)
        self.assertIsNone(roles.get_latest_archive())


class RepositoryTests(TempDirTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        logging.basicConfig(level=logging.DEBUG)

    def test_defaults(self):
        self.assertTrue(Repository(app_name='test'))

    def test_config_dict(self):
        app_name = 'test'
        repo = Repository(app_name=app_name)
        expected_config_dict = {
            'app_name': app_name,
            'app_version_attr': None,
            'encrypted_keys': None,
            'expiration_days': None,
            'key_map': None,
            'keys_dir': None,
            'repo_dir': None,
            'thresholds': None,
        }
        self.assertEqual(set(expected_config_dict), set(repo.config_dict))

    def test_app_version(self):
        # for convenience we use the notosotuf version attribute here,
        # but normally this would point to an external app version e.g.
        # 'my_app.__version__'
        app_version_attr = 'tufup.__version__'
        repo = Repository(app_name='test', app_version_attr=app_version_attr)
        self.assertEqual(str(tufup.__version__), repo.app_version)

    def test_get_config_file_path(self):
        self.assertTrue(Repository.get_config_file_path())

    def test_save_config(self):
        # prepare
        repo = Repository(app_name='test')
        # test
        repo.save_config()
        self.assertTrue(repo.get_config_file_path().exists())
        print(repo.get_config_file_path().read_text())

    def test_load_config(self):
        # file does not exist
        self.assertEqual(dict(), Repository.load_config())
        # file exists but invalid
        Repository.get_config_file_path().touch()
        # test
        self.assertEqual(dict(), Repository.load_config())

    def test_from_config(self):
        temp_dir = self.temp_dir_path.resolve()
        # prepare
        config_data = dict(
            app_name='test',
            app_version_attr='my_app.__version__',
            repo_dir=temp_dir / 'repo',
            keys_dir=temp_dir / 'keystore',
            key_map=dict(),
            encrypted_keys=[],
            expiration_days=dict(),
            thresholds=dict(),
        )
        Repository.get_config_file_path().write_text(
            json.dumps(config_data, default=str)
        )
        mock_load_keys_and_roles = Mock()
        # test
        with patch.object(
                Repository, '_load_keys_and_roles', mock_load_keys_and_roles,
        ):
            repo = Repository.from_config()
        self.assertEqual(
            config_data,
            {item: getattr(repo, item) for item in config_data.keys()},
        )
        self.assertTrue(mock_load_keys_and_roles.called)

    def test_initialize(self):
        # prepare
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
            expiration_days=DUMMY_EXPIRATION_DAYS,
        )
        # test
        repo.initialize()
        self.assertTrue(any(repo.keys_dir.iterdir()))
        self.assertTrue(any(repo.metadata_dir.iterdir()))
        self.assertTrue(
            all(getattr(repo.roles, name) for name in TOP_LEVEL_ROLE_NAMES)
        )
        self.assertEqual(
            date.today() + timedelta(days=DUMMY_EXPIRATION_DAYS['root']),
            repo.roles.root.signed.expires.date(),
        )

    def test_refresh_expiration_date(self):
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        days = 999
        repo.refresh_expiration_date(role_name='root', days=days)
        self.assertEqual(
            date.today() + timedelta(days=days),
            repo.roles.root.signed.expires.date(),
        )

    def test_replace_key(self):
        role_name = 'root'
        # prepare
        new_key_name = 'new_key'
        old_key_name = 'old-key'
        key_map = {name: [name] for name in TOP_LEVEL_ROLE_NAMES}
        key_map['root'].append(old_key_name)
        key_map['targets'] = [old_key_name]
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
            key_map=key_map,
        )
        repo.initialize()  # todo: make test independent...
        old_key_id = repo.roles.root.signed.roles['targets'].keyids[0]
        # create new key pair to replace old one
        new_private_key_path = repo.keys_dir / new_key_name
        new_public_key_path = Keys.create_key_pair(
            private_key_path=new_private_key_path, encrypted=False
        )
        new_key_id = import_ed25519_publickey_from_file(
            filepath=str(new_public_key_path)
        )['keyid']
        # test
        expected_key_count = len(key_map[role_name])
        repo.replace_key(
            old_key_name=old_key_name,
            new_public_key_path=new_public_key_path,
            new_private_key_encrypted=True,  # pretend the key is encrypted
        )
        self.assertEqual(
            expected_key_count,
            len(repo.roles.root.signed.roles[role_name].keyids),
        )
        # old key removed?
        self.assertNotIn(
            old_key_id, repo.roles.root.signed.roles[role_name].keyids
        )
        self.assertNotIn(old_key_name, repo.key_map[role_name])
        self.assertIn(old_key_name, repo.revoked_key_names)
        # new key added?
        self.assertIn(
            new_key_id, repo.roles.root.signed.roles[role_name].keyids
        )
        self.assertIn(new_key_name, repo.key_map[role_name])
        self.assertIn(new_key_name, repo.encrypted_keys)
        # no duplicates in encrypted_keys?
        self.assertEqual(1, len(repo.encrypted_keys))
        # other keys still in key map?
        self.assertEqual(expected_key_count, len(repo.key_map[role_name]))
        # keys replaced for all roles?
        self.assertIn(new_key_name, repo.key_map['targets'])

    def test_add_key(self):
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        # create new key pair to add
        new_key_name = 'new-key'
        new_private_key_path = repo.keys_dir / new_key_name
        new_public_key_path = Keys.create_key_pair(
            private_key_path=new_private_key_path, encrypted=False
        )
        new_key_id = import_ed25519_publickey_from_file(
            filepath=str(new_public_key_path)
        )['keyid']
        # test
        role_name = 'root'
        repo.add_key(
            role_name=role_name,
            public_key_path=new_public_key_path,
            encrypted=True,  # pretend the key is encrypted
        )
        # new key added?
        self.assertEqual(2, len(repo.roles.root.signed.roles[role_name].keyids))
        self.assertIn(
            new_key_id, repo.roles.root.signed.roles[role_name].keyids
        )
        self.assertIn(new_key_name, repo.key_map[role_name])
        self.assertIn(new_key_name, repo.encrypted_keys)

    def test_add_bundle(self):
        # prepare
        bundle_dir = self.temp_dir_path / 'dist' / 'test_app'
        bundle_dir.mkdir(parents=True)
        bundle_file = bundle_dir / 'dummy.exe'
        bundle_file.touch()
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        # test
        repo.add_bundle(new_version='1.0', new_bundle_dir=bundle_dir)
        self.assertTrue((repo.metadata_dir / 'targets.json').exists())

    def test_add_bundle_no_patch(self):
        # prepare
        bundle_dir = self.temp_dir_path / 'dist' / 'test_app'
        bundle_dir.mkdir(parents=True)
        bundle_file = bundle_dir / 'dummy.exe'
        bundle_file.write_text('this is version 1')
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        repo.add_bundle(new_version='1.0', new_bundle_dir=bundle_dir)
        # test
        bundle_file.write_text('much has changed in version 2')
        repo.add_bundle(
            new_version='2.0', new_bundle_dir=bundle_dir, skip_patch=True
        )
        self.assertTrue((repo.metadata_dir / 'targets.json').exists())
        target_keys = list(repo.roles.targets.signed.targets.keys())
        self.assertEqual(2, len(target_keys))
        self.assertFalse(any(key.endswith(SUFFIX_PATCH) for key in target_keys))

    def test_remove_latest_bundle(self):
        # prepare
        bundle_dir = self.temp_dir_path / 'dist' / 'test_app'
        bundle_dir.mkdir(parents=True)
        bundle_file = bundle_dir / 'dummy.exe'
        bundle_file.touch()
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        v1 = '1.0'
        v2 = '2.0'
        repo.add_bundle(new_version=v1, new_bundle_dir=bundle_dir)
        repo.add_bundle(new_version=v2, new_bundle_dir=bundle_dir)
        # test
        repo.remove_latest_bundle()
        remaining_target_keys = list(repo.roles.targets.signed.targets.keys())
        self.assertEqual(1, len(remaining_target_keys))
        self.assertIn(v1, remaining_target_keys[0])
        remaining_target_filenames = [p.name for p in repo.targets_dir.iterdir()]
        self.assertEqual(1, len(remaining_target_filenames))
        self.assertIn(remaining_target_keys[0], remaining_target_filenames)

    def test_threshold_sign(self):
        # prepare
        role_name = 'root'
        keys_dir = self.temp_dir_path / 'keystore'
        repo = Repository(
            app_name='test',
            keys_dir=keys_dir,
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        # create dummy "revoked key" and pretend it is encrypted
        revoked_key_name = 'revoked'
        repo.keys.create_key_pair(
            private_key_path=keys_dir / revoked_key_name, encrypted=False
        )
        repo.revoked_key_names.append(revoked_key_name)
        repo.encrypted_keys.append(revoked_key_name)
        # dummy modification
        repo.roles.root.signed.spec_version = 'x'
        versioned_file_path = repo.metadata_dir / f'1.{role_name}.json'
        non_versioned_file_path = repo.metadata_dir / f'{role_name}.json'
        versioned_last_modified = versioned_file_path.stat().st_mtime_ns
        non_versioned_last_modified = non_versioned_file_path.stat().st_mtime_ns
        # test
        sleep(0.1)  # enforce different modification time
        count = repo.threshold_sign(
            role_name=role_name, private_key_dirs=[repo.keys_dir]
        )
        self.assertEqual(2, count)  # existing key and revoked key
        self.assertFalse(repo.revoked_key_names)
        self.assertFalse(repo.encrypted_keys)
        # files should have been modified
        self.assertGreater(
            versioned_file_path.stat().st_mtime_ns, versioned_last_modified
        )
        self.assertGreater(
            non_versioned_file_path.stat().st_mtime_ns, non_versioned_last_modified
        )
        # if no signatures found
        for path in repo.keys_dir.iterdir():
            path.unlink()
        with self.assertRaises(Exception):
            repo.threshold_sign(
                role_name=role_name, private_key_dirs=[repo.keys_dir]
            )

    def test__load_keys_and_roles(self):
        # prepare
        keys_dir = self.temp_dir_path / 'keystore'
        repo = Repository(
            app_name='test',
            keys_dir=keys_dir,
            repo_dir=self.temp_dir_path / 'repo',
        )
        # test
        with patch('builtins.input', Mock(return_value='y')):
            repo._load_keys_and_roles(create_keys=True)
        self.assertTrue(
            all(getattr(repo.roles, name) for name in TOP_LEVEL_ROLE_NAMES)
        )
        self.assertTrue(
            all((keys_dir / name).exists() for name in TOP_LEVEL_ROLE_NAMES)
        )

    def test_publish_changes_no_change(self):
        # prepare
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        # note that initialize() already calls publish_changes...
        repo.initialize()  # todo: make test independent...
        # test
        repo.publish_changes(private_key_dirs=[repo.keys_dir])
        for role_name in ['root', 'targets', 'snapshot', 'timestamp']:
            role = getattr(repo.roles, role_name)
            with self.subTest(msg=role_name):
                self.assertEqual(1, role.signed.version)
                self.assertTrue(role.signatures)

    def test_publish_changes_threshold(self):
        # prepare
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        # note that initialize() already calls publish_changes...
        repo.initialize()  # todo: make test independent...
        # remove the root signature
        repo.roles.root.signatures = {}
        # test
        repo.publish_changes(private_key_dirs=[repo.keys_dir])
        for role_name in ['root', 'targets', 'snapshot', 'timestamp']:
            with self.subTest(msg=role_name):
                role = getattr(repo.roles, role_name)
                self.assertEqual(
                    repo.thresholds[role_name], len(role.signatures)
                )

    def test_publish_changes(self):
        days = 999
        role_names = ['root', 'targets', 'snapshot', 'timestamp']
        for test_role_name in role_names.copy():
            # use a fresh temporary dir and repo
            with TemporaryDirectory() as temp_dir:
                temp_dir_path = pathlib.Path(temp_dir)
                repo = Repository(
                    app_name='test',
                    keys_dir=temp_dir_path / 'keystore',
                    repo_dir=temp_dir_path / 'repo',
                )
                # note that initialize() already calls publish_changes...
                repo.initialize()  # todo: make test independent...
                # make a change to metadata (in memory)
                repo.roles.set_expiration_date(
                    role_name=test_role_name, days=days
                )
                # make a change to config
                config_change = 'dummy'
                repo.encrypted_keys.append(config_change)
                # test
                repo.publish_changes(private_key_dirs=[repo.keys_dir])
                for role_name in role_names:
                    role = getattr(repo.roles, role_name)
                    with self.subTest(msg=f'{test_role_name}: {role_name}'):
                        self.assertEqual(2, role.signed.version)
                        self.assertTrue(role.signatures)
                role_names.remove(test_role_name)
                #  re-load from file to verify change
                root = Metadata.from_file(
                    repo.roles.file_path(role_name=test_role_name)
                )
                self.assertEqual(
                    date.today() + timedelta(days=days),
                    root.signed.expires.date(),
                )
                # verify change in config
                config_from_disk = repo.load_config()
                self.assertIn(
                    config_change, config_from_disk['encrypted_keys']
                )



import argparse
import logging

import packaging.version
from tuf.api.metadata import TOP_LEVEL_ROLE_NAMES

from tufup.utils import (
    log_print, input_bool, input_numeric, input_text, input_list
)
from tufup.repo import Repository

logger = logging.getLogger(__name__)

HELP = dict(
    common_key_dirs='Directories to search for private and/or public keys.',
    targets_add=(
        'Add app bundle to the repository. Creates archive and patch from'
        ' bundle files.'
    ),
    targets_add_app_version='Application version (PEP440 compliant)',
    targets_add_bundle_dir='Directory containing application bundle.',
    targets_add_skip_patch='Skip patch creation.',
    targets_remove_latest='Remove latest app bundle from the repository.',
    keys_subcommands='Optional commands to add or replace keys.',
    keys_new_key_name='Name of new private key (public key gets .pub suffix).',
    keys_role_name='Register public key for this role.',
    keys_encrypted='New private key is (to be) encrypted.',
    keys_create='Create a new key pair (private & public).',
    keys_old_key_name='Revoke old public key, replace by new public key.',
    sign_role_name='Name of role to be signed.',
    sign_expiration_days=(
        'Set expiration date as number of days from today. Metadata version '
        'and expiration date for dependent roles will also be updated.'
    ),
)


def _print_info(message: str):
    return log_print(message=message, level=logging.INFO, logger=logger)


def _get_repo():
    try:
        return Repository.from_config()
    except TypeError:
        _print_info(
            'Failed to load config. Did you initialize the repository?'
        )


def _add_key_dirs_argument(parser: argparse.ArgumentParser):
    parser.add_argument('key_dirs', nargs='+', help=HELP['common_key_dirs'])


def get_parser() -> argparse.ArgumentParser:
    # https://docs.python.org/3/library/argparse.html#sub-commands
    # https://docs.python.org/3/library/argparse.html#parents
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    # add debug option
    debug_parser = argparse.ArgumentParser(add_help=False)
    debug_parser.add_argument(
        '-d', '--debug', action='store_true', required=False
    )
    # init
    subparser_init = subparsers.add_parser('init', parents=[debug_parser])
    subparser_init.set_defaults(func=_cmd_init)
    # targets
    subparser_targets = subparsers.add_parser('targets', parents=[debug_parser])
    subparser_targets.set_defaults(func=_cmd_targets)
    # we use nested subparsers to deal with mutually dependent arguments
    targets_subparsers = subparser_targets.add_subparsers()
    subparser_targets_add = targets_subparsers.add_parser(
        'add', help=HELP['targets_add']
    )
    subparser_targets_add.add_argument(
        'app_version',
        action=_StoreVersionAction,
        help=HELP['targets_add_app_version'],
    )
    subparser_targets_add.add_argument(
        'bundle_dir', help=HELP['targets_add_bundle_dir']
    )
    subparser_targets_add.add_argument(
        '-s',
        '--skip-patch',
        action='store_true',
        required=False,
        help=HELP['targets_add_skip_patch'],
    )
    subparser_targets_remove = targets_subparsers.add_parser(
        'remove-latest', help=HELP['targets_remove_latest']
    )
    for sp in [subparser_targets_add, subparser_targets_remove]:
        _add_key_dirs_argument(parser=sp)
    # keys
    subparser_keys = subparsers.add_parser('keys', parents=[debug_parser])
    subparser_keys.set_defaults(func=_cmd_keys)
    subparser_keys.add_argument(
        'new_key_name', help=HELP['keys_new_key_name']
    )
    subparser_keys.add_argument(
        '-c', '--create', action='store_true', help=HELP['keys_create']
    )
    subparser_keys.add_argument(
        '-e', '--encrypted', action='store_true', help=HELP['keys_encrypted']
    )
    # we use nested subparsers to deal with mutually dependent arguments
    keys_subparsers = subparser_keys.add_subparsers(
        help=HELP['keys_subcommands']
    )
    subparser_keys_add = keys_subparsers.add_parser('add')
    subparser_keys_add.add_argument(
        'role_name', choices=TOP_LEVEL_ROLE_NAMES, help=HELP['keys_role_name']
    )
    subparser_keys_replace = keys_subparsers.add_parser('replace')
    subparser_keys_replace.add_argument(
        'old_key_name', help=HELP['keys_old_key_name']
    )
    for sp in [subparser_keys_add, subparser_keys_replace]:
        _add_key_dirs_argument(parser=sp)
    # sign
    subparser_sign = subparsers.add_parser('sign', parents=[debug_parser])
    subparser_sign.set_defaults(func=_cmd_sign)
    subparser_sign.add_argument(
        'role_name', choices=TOP_LEVEL_ROLE_NAMES, help=HELP['sign_role_name']
    )
    subparser_sign.add_argument(
        '-e',
        '--expiration-days',
        required=False,
        nargs='?',
        const='default',  # if option -e is specified without value
        default=None,  # if option -e is not specified at all
        help=HELP['sign_expiration_days'],
    )
    _add_key_dirs_argument(parser=subparser_sign)
    return parser


class _StoreVersionAction(argparse.Action):
    """Validates version string before storing."""
    def __call__(self, parser, namespace, values, option_string=None, **kwargs):
        # The first value should comply with PEP440
        value = values[0]
        try:
            packaging.version.Version(value)
        except packaging.version.InvalidVersion:
            raise argparse.ArgumentError(
                self,
                f'Version string "{value}" is not PEP440 compliant.\n '
                f'See examples: https://www.python.org/dev/peps/pep-0440/\n'
            )
        # Store the value, same as "store" action
        setattr(namespace, self.dest, values)


def _get_config_from_user(**kwargs) -> dict:
    top_level_role_names = ['root', 'targets', 'snapshot', 'timestamp']
    for key, example, optional in [
        ('app_name', '', False),
        ('app_version_attr', ', e.g. my_app.__version__', True),
        ('repo_dir', '', False),
        ('keys_dir', '', False),
    ]:
        kwargs[key] = input_text(
            prompt=f'Specify {key}{example}',
            default=kwargs.get(key),
            optional=optional,
        )
    key_map = kwargs.get('key_map', {})
    expiration_days = kwargs.get('expiration_days', {})
    thresholds = kwargs.get('thresholds', {})
    encrypted_keys = kwargs.get('encrypted_keys', [])
    new_encrypted_keys = []
    unique_key_names = []
    for role_name in top_level_role_names:
        # key_map
        key_names = []
        while not key_names:
            key_names = input_list(
                prompt=f'Specify key names for {role_name}',
                default=key_map.get(role_name, [role_name]),
            )
        key_map[role_name] = key_names
        # encrypted_keys
        for key_name in key_names:
            if key_name not in unique_key_names:
                unique_key_names.append(key_name)
                if input_bool(
                        prompt=f'Encrypt key "{key_name}"?',
                        default=key_name in encrypted_keys,
                ):
                    new_encrypted_keys.append(key_name)
        # expiration_days
        expiration_days[role_name] = input_numeric(
            prompt=f'Specify number of days before {role_name} expires',
            default=expiration_days.get(role_name, 1),
        )
        # thresholds
        thresholds[role_name] = input_numeric(
            prompt=f'Specify required number of signatures for {role_name}',
            default=thresholds.get(role_name, 1),
        )
    kwargs['key_map'] = key_map
    kwargs['expiration_days'] = expiration_days
    kwargs['encrypted_keys'] = new_encrypted_keys
    kwargs['thresholds'] = thresholds
    return kwargs


def _cmd_init(options: argparse.Namespace):
    logger.debug(f'command init: {vars(options)}')
    # try to load existing config
    config_dict = Repository.load_config()
    modify = True
    message = 'Creating new configuration.'
    if config_dict:
        modify = input_bool(
            prompt='Found existing configuration. Modify?', default=False
        )
        if modify:
            message = 'Modifying existing configuration.'
        else:
            message = 'Using existing configuration.'
    _print_info(message)
    if modify:
        config_dict = _get_config_from_user(**config_dict)
    # create repository instance
    repository = Repository(**config_dict)
    # save new or updated configuration
    _print_info('Saving config...')
    repository.save_config()
    # create directories, keys, and root metadata file
    _print_info('Initializing repository...')
    repository.initialize()
    _print_info('Done.')


def _cmd_keys(options: argparse.Namespace):
    logger.debug(f'command keys: {vars(options)}')
    repository = _get_repo()
    public_key_path = repository.keys.public_key_path(
        key_name=options.new_key_name
    )
    private_key_path = repository.keys.private_key_path(
        key_name=options.new_key_name
    )
    if options.create:
        _print_info(f'Creating key pair for {options.new_key_name}...')
        repository.keys.create_key_pair(
            private_key_path=private_key_path, encrypted=options.encrypted
        )
        _print_info(f'Key pair created.')
    replace = hasattr(options, 'old_key_name')
    add = hasattr(options, 'role_name')
    if replace:
        _print_info(
            f'Replacing key {options.old_key_name} by {options.new_key_name}...'
        )
        repository.replace_key(
            old_key_name=options.old_key_name,
            new_public_key_path=public_key_path,
            new_private_key_encrypted=options.encrypted,
        )
    elif add:
        _print_info(f'Adding key {options.new_key_name}...')
        repository.add_key(
            role_name=options.role_name,
            public_key_path=public_key_path,
            encrypted=options.encrypted,
        )
    if replace or add:
        _print_info('Publishing changes...')
        repository.publish_changes(private_key_dirs=options.key_dirs)
    _print_info('Done.')


def _cmd_targets(options: argparse.Namespace):
    logger.debug(f'command targets: {vars(options)}')
    repository = _get_repo()
    if hasattr(options, 'app_version') and hasattr(options, 'bundle_dir'):
        _print_info('Adding bundle...')
        repository.add_bundle(
            new_version=options.app_version,
            new_bundle_dir=options.bundle_dir,
            skip_patch=options.skip_patch,
        )
    else:
        _print_info('Removing latest bundle...')
        repository.remove_latest_bundle()
    _print_info('Publishing changes...')
    repository.publish_changes(private_key_dirs=options.key_dirs)
    _print_info('Done.')


def _cmd_sign(options: argparse.Namespace):
    logger.debug(f'command sign: {vars(options)}')
    repository = _get_repo()
    if options.expiration_days is not None:
        # default or custom
        days = repository.expiration_days.get(options.role_name)
        if options.expiration_days.isnumeric():
            days = int(options.expiration_days)
        # change expiration date in signed metadata
        _print_info(f'Setting expiration date {days} days from now...')
        repository.refresh_expiration_date(
            role_name=options.role_name, days=days
        )
        # also update version and expiration date for dependent roles, and sign
        # modified roles
        _print_info('Publishing changes...')
        repository.publish_changes(private_key_dirs=options.key_dirs)
    else:
        # sign without changing the signed metadata (for threshold signing)
        _print_info('Adding signature...')
        repository.threshold_sign(
            role_name=options.role_name,
            private_key_dirs=options.key_dirs,
        )
    _print_info('Done.')


import logging
import os
import pathlib
import shutil
from typing import Optional
import unittest
from unittest.mock import Mock, patch

from requests.auth import HTTPBasicAuth
import tuf.api.exceptions
from tuf.ngclient import TargetFile

from tests import TempDirTestCase, TEST_REPO_DIR
from tufup.client import AuthRequestsFetcher, Client
from tufup.common import TargetMeta

ROOT_FILENAME = 'root.json'
TARGETS_FILENAME = 'targets.json'
SNAPSHOT_FILENAME = 'snapshot.json'
TIMESTAMP_FILENAME = 'timestamp.json'
ON_GITHUB = os.getenv('GITHUB_ACTIONS')


class ClientTests(TempDirTestCase):
    def setUp(self) -> None:
        super().setUp()
        # directory where the parent application is installed (e.g.
        # %PROGRAMFILES%\MyApp or %LOCALAPPDATA%\Programs\MyApp on Windows 10)
        # https://docs.microsoft.com/en-us/windows/win32/msi/installation-context
        self.app_install_dir = self.temp_dir_path / 'programs' / 'example'
        self.app_install_dir.mkdir(parents=True)
        # directories must be created by parent application
        self.metadata_dir = self.temp_dir_path / 'metadata'
        self.target_dir = self.temp_dir_path / 'targets'
        self.metadata_dir.mkdir()
        self.target_dir.mkdir()
        # parent application must be shipped with root metadata, and must
        # ensure it is placed in the metadata_dir (without version in filename)
        shutil.copy(
            src=TEST_REPO_DIR / 'metadata' / ('1.' + ROOT_FILENAME),
            dst=self.metadata_dir / ROOT_FILENAME,
        )
        # kwargs for client initializer
        self.client_kwargs = dict(
            app_name='example_app',
            app_install_dir=self.app_install_dir,
            current_version='1.0',
            metadata_dir=self.metadata_dir,
            metadata_base_url='http://localhost:8000/metadata/',
            target_dir=self.target_dir,
            target_base_url='http://localhost:8000/targets/',
            session_auth={'http://localhost:8000': ('username', 'password')}
        )

    def mock_download_metadata(
            self, rolename: str, length: int, version: Optional[int] = None
    ) -> bytes:
        if rolename == 'root':
            # indicate current root is newest version
            raise tuf.api.exceptions.DownloadHTTPError(status_code=404, message='')
        # read from the test repo dir, instead of actually downloading
        filename = f'{rolename}.json'
        if version:
            filename = f'{version}.{filename}'
        file_path = TEST_REPO_DIR / 'metadata' / filename
        return file_path.read_bytes()

    def get_refreshed_client(self):
        # refresh to load targets metadata (mock to prevent actual download)
        client = Client(**self.client_kwargs)
        with patch.object(client, '_download_metadata', self.mock_download_metadata):
            client.refresh()
        # ensure current archive exists (dummy)
        shutil.copy(
            src=TEST_REPO_DIR / 'targets' / client.current_archive.path.name,
            dst=client.current_archive_local_path,
        )
        return client

    def test_init_no_metadata(self):
        # cannot initialize without root metadata file
        (self.metadata_dir / ROOT_FILENAME).unlink()
        with self.assertRaises(FileNotFoundError):
            Client(**self.client_kwargs)

    def test_init(self):
        client = Client(**self.client_kwargs)
        # Client is a subclass of tuf.ngclient.Updater, so it automatically
        # loads and verifies the local root metadata file
        self.assertTrue(client._trusted_set.root)
        # other metadata is not available yet
        for role_name in ['targets', 'snapshot', 'timestamp']:
            # see python-tuf #2250
            self.assertNotIn(role_name, client._trusted_set)

    def test_trusted_target_metas(self):
        client = self.get_refreshed_client()
        self.assertTrue(client.trusted_target_metas)

    def test_get_targetinfo(self):
        client = self.get_refreshed_client()
        target_path_str = 'example_app-1.0.tar.gz'
        target_meta = TargetMeta(target_path=target_path_str)
        for target_path in [target_path_str, target_meta]:
            target_info = client.get_targetinfo(target_path=target_path)
            with self.subTest(msg=target_path):
                self.assertIsInstance(target_info, TargetFile)

    def test_updates_available(self):
        client = Client(**self.client_kwargs)
        # test check_for_updates not called
        with self.assertLogs(logger='tufup.client', level=logging.WARNING) as cm:
            self.assertFalse(client.updates_available)
        self.assertIn('check_for_updates', cm.output[0])
        # test check_for_updates called and update available
        client.new_targets = {'dummy': None}
        self.assertTrue(client.updates_available)

    def test_download_and_apply_update(self):
        # just for completeness...
        mock_download = Mock(return_value=True)
        mock_apply = Mock(return_value=True)
        mock_install = Mock()
        with patch.multiple(
                Client, _download_updates=mock_download, _apply_updates=mock_apply
        ):
            client = self.get_refreshed_client()
            client.new_targets = {'dummy': None}
            client.download_and_apply_update(install=mock_install)
        self.assertTrue(mock_download.called)
        self.assertTrue(mock_apply.called)
        self.assertIn(mock_install, mock_apply.call_args.kwargs.values())

    def test_check_for_updates(self):
        # expectations (based on targets in tests/data/repository):
        # - pre=None: only full releases are included, so finds 2.0 patch
        # - pre='a': finds all, but total patch size exceeds archive size
        # - pre='b': there is no 'b' release, so this finds same as 'rc'
        # - pre='rc': finds 2.0 and 3.0rc0, total patch size smaller than archive
        client = self.get_refreshed_client()
        with patch.object(client, 'refresh', Mock()):
            for pre, expected in [(None, 1), ('a', 1), ('b', 2), ('rc', 2)]:
                with self.subTest(msg=pre):
                    self.assertTrue(client.check_for_updates(pre=pre))
                    self.assertEqual(expected, len(client.new_targets))
                    if pre == 'a':
                        self.assertTrue(all(
                            item.is_archive for item in
                            client.new_targets.keys())
                        )
                    else:
                        self.assertTrue(all(
                            item.is_patch for item in client.new_targets.keys())
                        )

    def test_check_for_updates_already_up_to_date(self):
        self.client_kwargs['current_version'] = '4.0a0'
        client = self.get_refreshed_client()
        with patch.object(client, 'refresh', Mock()):
            self.assertFalse(client.check_for_updates(pre='a'))

    def test_check_for_updates_current_archive_missing(self):
        client = self.get_refreshed_client()
        # remove current archive dummy
        client.current_archive_local_path.unlink()
        with patch.object(client, 'refresh', Mock()):
            for pre in [None, 'a', 'b', 'rc']:
                self.assertTrue(client.check_for_updates(pre=pre))
                target_meta = next(iter(client.new_targets.keys()))
                self.assertTrue(target_meta.is_archive)

    def test__download_updates(self):
        client = Client(**self.client_kwargs)
        client.new_targets = {Mock(): Mock()}
        for cached_path, downloaded_path in [('cached', None), (None, 'downloaded')]:
            with patch.multiple(
                    client,
                    find_cached_target=Mock(return_value=cached_path),
                    download_target=Mock(return_value=downloaded_path),
            ):
                self.assertTrue(client._download_updates(progress_hook=None))
                local_path = next(iter(client.downloaded_target_files.values()))
                if cached_path:
                    self.assertEqual(cached_path, str(local_path))
                else:
                    self.assertEqual(downloaded_path, str(local_path))

    def test__apply_updates(self):
        client = self.get_refreshed_client()
        # directly use target files from test repo as downloaded files
        client.downloaded_target_files = {
            target_meta: TEST_REPO_DIR / 'targets' / str(target_meta)
            for target_meta in client.trusted_target_metas
            if target_meta.is_patch
            and str(target_meta.version) in ['2.0', '3.0rc0']
        }
        # specify new archive (normally done in _check_updates)
        archives = [
            tp for tp in client.trusted_target_metas
            if tp.is_archive and str(tp.version) == '3.0rc0'
        ]
        client.new_archive_info = client.get_targetinfo(archives[-1])
        client.new_archive_local_path = pathlib.Path(
            client.target_dir, client.new_archive_info.path
        )
        # test confirmation
        mock_install = Mock()
        with patch('builtins.input', Mock(return_value='y')):
            client._apply_updates(install=mock_install, skip_confirmation=False)
        self.assertTrue(any(client.extract_dir.iterdir()))
        self.assertTrue(mock_install.called)
        # test skip confirmation
        mock_install = Mock()
        client._apply_updates(install=mock_install, skip_confirmation=True)
        mock_install.assert_called()


class AuthRequestsFetcherTests(unittest.TestCase):
    def setUp(self) -> None:
        self.session_auth = {
            'https://example.net': None,
            'https://example.com': ('username', 'password'),
            'https://example.org': HTTPBasicAuth(username='x', password='y'),
            'http://localhost:8000': ('username', 'password'),
        }

    def test_init(self):
        # drop-in replacement for default RequestsFetcher, without args
        self.assertTrue(AuthRequestsFetcher())
        # if authentication is required, specify arg
        fetcher = AuthRequestsFetcher(session_auth=self.session_auth)
        self.assertEqual(self.session_auth, fetcher.session_auth)

    def test__get_session(self):
        fetcher = AuthRequestsFetcher(session_auth=self.session_auth)
        for scheme_and_server, auth in self.session_auth.items():
            url = scheme_and_server + '/some/path?query=something'
            with self.subTest(msg=url):
                session = fetcher._get_session(url=url)
                self.assertEqual(auth, session.auth)

    @unittest.skipIf(condition=ON_GITHUB, reason='external dependency')
    def test_fetch_basic_auth(self):
        # kind of an integration test, as it connects to an external server...
        scheme_and_server = 'https://httpbin.org'
        user = 'me'
        passwd = 'mypassword'
        url = f'{scheme_and_server}/basic-auth/{user}/{passwd}'
        session_auth = {
            scheme_and_server: HTTPBasicAuth(username=user, password=passwd)
        }
        fetcher = AuthRequestsFetcher(session_auth=session_auth)
        fetcher.socket_timeout = 30  # in case httpbin.org is slow to respond
        # we don't have direct access to the response, so we'll just check
        # that RequestsFetcher.fetch() doesn't raise an error, such as a
        # status "401 Unauthorized" or "403 Forbidden"
        try:
            fetcher.fetch(url=url)
        except tuf.api.exceptions.DownloadHTTPError as e:
            self.fail(msg=f'fetch() raised unexpected HTTPError: {e}')

    def test_attach_progress_hook(self):
        mock_hook = Mock()
        bytes_expected = 10
        fetcher = AuthRequestsFetcher()
        fetcher.attach_progress_hook(
            hook=mock_hook, bytes_expected=bytes_expected
        )
        bytes_new = 1
        bytes_downloaded = 0
        while bytes_downloaded < bytes_expected:
            bytes_downloaded += bytes_new
            fetcher._progress(bytes_new=bytes_new)
            mock_hook.assert_called_with(
                bytes_downloaded=bytes_downloaded, bytes_expected=bytes_expected
            )

    def test__chunks_without_progress_hook(self):
        chunk_size = 10
        chunk_count = 10
        chunks = [b'x' * chunk_size] * chunk_count

        def mock_iter_content(*args):
            yield from chunks

        mock_response = Mock(iter_content=mock_iter_content, close=Mock())
        fetcher = AuthRequestsFetcher()
        fetcher.chunk_size = chunk_size
        # _chunks should work even if attach_progress_hook was not called
        try:
            for __ in fetcher._chunks(response=mock_response):
                pass
        except Exception as e:
            self.fail(msg=f'_chunks raised an unexpected exception: {e}')

    def test__chunks_with_progress_hook(self):
        chunk_size = 10
        chunk_count = 10
        chunks = [b'x' * chunk_size] * chunk_count

        def mock_iter_content(*args):
            yield from chunks

        mock_response = Mock(iter_content=mock_iter_content, close=Mock())
        fetcher = AuthRequestsFetcher()
        fetcher.chunk_size = chunk_size
        # test custom progress hook
        mock_hook = Mock()
        bytes_expected = chunk_size * chunk_count
        fetcher.attach_progress_hook(
            hook=mock_hook, bytes_expected=bytes_expected
        )
        for __ in fetcher._chunks(response=mock_response):
            pass
        self.assertEqual(chunk_count, mock_hook.call_count)



from copy import deepcopy
from datetime import datetime, timedelta
import inspect
import json
import logging
import pathlib
try:
    # workaround for PyInstaller issue 6911 (setuptools issue 3089)
    import setuptools.config.expand  # noqa
except AssertionError as e:
    pass  # assuming we are on the client side...
import shutil
from typing import Any, Dict, Iterable, List, Optional, TypedDict, Union

from securesystemslib.exceptions import CryptoError
from securesystemslib.interface import (
    generate_and_write_ed25519_keypair_with_prompt,
    generate_and_write_unencrypted_ed25519_keypair,
    import_ed25519_publickey_from_file,
    import_ed25519_privatekey_from_file,
)
# SSlibKey see: https://github.com/secure-systems-lab/securesystemslib/pull/456
from securesystemslib.signer import SSlibKey, SSlibSigner
from tuf.api.metadata import (
    SPECIFICATION_VERSION,
    TOP_LEVEL_ROLE_NAMES,
    Metadata,
    MetaFile,
    Role,
    Root,
    Snapshot,
    TargetFile,
    Targets,
    Timestamp,
)
from tuf.api.serialization.json import JSONSerializer

from tufup.common import Patcher, SUFFIX_ARCHIVE, SUFFIX_PATCH, TargetMeta

logger = logging.getLogger(__name__)

"""

https://github.com/theupdateframework/python-tuf/blob/develop/examples/repo_example/basic_repo.py

"""

__all__ = [
    'DEFAULT_KEY_MAP',
    'DEFAULT_KEYS_DIR_NAME',
    'DEFAULT_META_DIR_NAME',
    'DEFAULT_REPO_DIR_NAME',
    'DEFAULT_TARGETS_DIR_NAME',
    'in_',
    'Keys',
    'make_gztar_archive',
    'Repository',
    'Roles',
    'SUFFIX_JSON',
    'SUFFIX_PUB',
    'TOP_LEVEL_ROLE_NAMES',
]

# copied from python-tuf basic_repo.py
SPEC_VERSION = ".".join(SPECIFICATION_VERSION)


# copied from python-tuf basic_repo.py
def in_(days: float) -> datetime:
    """Returns a timestamp for the specified number of days from now."""
    return datetime.utcnow().replace(microsecond=0) + timedelta(days=days)


def make_gztar_archive(
        src_dir: Union[pathlib.Path, str],
        dst_dir: Union[pathlib.Path, str],
        app_name: str,
        version: str,
        **kwargs,  # allowed kwargs are passed on to shutil.make_archive
) -> Optional[TargetMeta]:
    # remove disallowed kwargs
    for key in ['base_name', 'root_dir', 'format']:
        if kwargs.pop(key, None):
            logger.warning(f'{key} ignored: using default')
    # ensure paths
    src_dir = pathlib.Path(src_dir)
    dst_dir = pathlib.Path(dst_dir)
    # compose archive path and check existence
    archive_filename = TargetMeta.compose_filename(
        name=app_name, version=version, is_archive=True
    )
    archive_path = dst_dir / archive_filename
    if archive_path.exists():
        if input(f'Found existing archive: {archive_path}.\nOverwrite? [n]/y') != 'y':
            print('Using existing archive.')
            return TargetMeta(archive_path)
    # make archive
    base_name = str(dst_dir / archive_filename.replace(SUFFIX_ARCHIVE, ''))
    archive_path_str = shutil.make_archive(
        base_name=base_name,  # archive file path, no suffix
        root_dir=str(src_dir),  # paths in archive will be relative to root_dir
        format='gztar',
        **kwargs,
    )
    return TargetMeta(target_path=archive_path_str)


class RolesDict(TypedDict):
    root: Any
    targets: Any
    snapshot: Any
    timestamp: Any


DEFAULT_REPO_DIR_NAME = 'repository'
DEFAULT_KEYS_DIR_NAME = 'keystore'
DEFAULT_META_DIR_NAME = 'metadata'
DEFAULT_TARGETS_DIR_NAME = 'targets'
DEFAULT_KEY_MAP = RolesDict((key, [key]) for key in TOP_LEVEL_ROLE_NAMES)  # noqa
DEFAULT_EXPIRATION_DAYS = RolesDict(root=365, targets=7, snapshot=7, timestamp=1)
DEFAULT_THRESHOLDS = RolesDict(root=1, targets=1, snapshot=1, timestamp=1)
SUFFIX_JSON = '.json'
SUFFIX_PUB = '.pub'
FILENAME_ROOT = Root.type + SUFFIX_JSON
FILENAME_TARGETS = Targets.type + SUFFIX_JSON
FILENAME_SNAPSHOT = Snapshot.type + SUFFIX_JSON
FILENAME_TIMESTAMP = Timestamp.type + SUFFIX_JSON


class Base(object):
    def __init__(self, dir_path: Union[pathlib.Path, str, None]):
        """
        dir_path: directory where all key files are stored
        encrypted: names of the keys that are (to be) encrypted
        key_map: maps top-level role names to lists of key names
        """
        if dir_path is None:
            dir_path = pathlib.Path.cwd()
        # enforce pathlib.Path
        self.dir_path = pathlib.Path(dir_path)
        if not self.dir_path.exists():
            if input(f'Create directory {self.dir_path}? [y]/n') in ['y', '']:
                self.dir_path.mkdir(parents=True)
                print(f'Directory created: {self.dir_path}')


class Keys(Base):
    filename_pattern = '{key_name}'

    def __init__(
            self,
            dir_path: Union[pathlib.Path, str, None] = None,
            encrypted: Optional[List[str]] = None,
            key_map: Optional[RolesDict] = None,
            thresholds: Optional[RolesDict] = None,
    ):
        if dir_path is None:
            dir_path = pathlib.Path.cwd() / DEFAULT_KEYS_DIR_NAME
        super().__init__(dir_path=dir_path)
        if encrypted is None:
            encrypted = []
        if key_map is None:
            key_map = deepcopy(DEFAULT_KEY_MAP)
        if thresholds is None:
            thresholds = DEFAULT_THRESHOLDS.copy()
        self.encrypted = encrypted
        self.key_map = key_map
        self.thresholds = thresholds
        # top-level roles
        self.root: List[Dict[str, Any]] = []
        self.targets: List[Dict[str, Any]] = []
        self.snapshot: List[Dict[str, Any]] = []
        self.timestamp: List[Dict[str, Any]] = []
        # import public keys from dir_path, if it exists
        self.import_all_public_keys()

    def import_all_public_keys(self):
        for role_name, key_list in self.key_map.items():
            for key_name in key_list:
                self.import_public_key(role_name=role_name, key_name=key_name)

    def import_public_key(self, role_name: str, key_name: Optional[str] = None):
        """Import public key for specified role."""
        if key_name is None:
            key_name = role_name
        public_key_path = self.public_key_path(key_name=key_name)
        if public_key_path.exists():
            ssl_key = import_ed25519_publickey_from_file(
                filepath=str(public_key_path)
            )
            getattr(self, role_name).append(ssl_key)
            logger.debug(f'public key imported: {public_key_path}')
        else:
            logger.debug(f'file does not exist: {public_key_path}')

    def create(self):
        all_key_names = []
        for key_list in self.key_map.values():
            all_key_names.extend(key_list)
        unique_key_names = set(all_key_names)
        logger.debug(f'creating key-pairs: {unique_key_names}')
        for key_name in unique_key_names:
            default_private_key_path = self.private_key_path(key_name=key_name)
            self.create_key_pair(
                private_key_path=default_private_key_path,
                encrypted=key_name in self.encrypted,
            )
        # import the newly created public keys
        self.import_all_public_keys()

    @staticmethod
    def create_key_pair(
            private_key_path: pathlib.Path, encrypted: bool
    ) -> pathlib.Path:
        if encrypted:
            # encrypt private key
            logger.debug(f'set encryption password for private key')
            generate_keypair = generate_and_write_ed25519_keypair_with_prompt
        else:
            # do not encrypt private key (for automated signing)
            generate_keypair = generate_and_write_unencrypted_ed25519_keypair
        public_key_path = private_key_path.with_suffix(SUFFIX_PUB)
        proceed = True
        if public_key_path.exists():
            logger.warning(f'Public key already exists: {public_key_path}')
            proceed = input(f'Overwrite key pair? [n]/y') == 'y'
        if proceed:
            file_path_str = generate_keypair(filepath=str(private_key_path))
            logger.info(f'key-pair created: {file_path_str}, {public_key_path}')
        return public_key_path

    def private_key_path(self, key_name: str) -> pathlib.Path:
        return self.dir_path / self.filename_pattern.format(key_name=key_name)

    def public_key_path(self, key_name: str) -> pathlib.Path:
        return self.private_key_path(key_name=key_name).with_suffix(SUFFIX_PUB)

    def public(self):
        # return a dict that maps key ids to *public* key objects
        return {
            ssl_key['keyid']: SSlibKey.from_securesystemslib_key(key_dict=ssl_key)
            for attr_name, ssl_keys in vars(self).items()
            if attr_name in TOP_LEVEL_ROLE_NAMES
            for ssl_key in ssl_keys
        }

    def roles(self):
        # return a dict that maps role names to key ids and key thresholds
        roles_map = dict()
        for role_name in TOP_LEVEL_ROLE_NAMES:
            ssl_keys = getattr(self, role_name)
            role_keys = None
            if ssl_keys:
                unique_key_ids = list(set(ssl_key['keyid'] for ssl_key in ssl_keys))
                role_keys = Role(
                    keyids=unique_key_ids, threshold=self.thresholds[role_name]
                )
            roles_map[role_name] = role_keys
        return roles_map

    @classmethod
    def find_private_key(cls, key_name: str, key_dirs: List[Union[pathlib.Path, str]]):
        private_key_path = None
        private_key_filename = cls.filename_pattern.format(key_name=key_name)
        for key_dir in key_dirs:
            key_dir = pathlib.Path(key_dir)  # ensure Path
            for path in key_dir.iterdir():
                if path.is_file() and path.name == private_key_filename:
                    private_key_path = path
                    break
        return private_key_path


class Roles(Base):
    filename_pattern = '{version}{role_name}{suffix}'

    def __init__(self, dir_path: Union[pathlib.Path, str, None] = None):
        """
        TUF roles

        - root metadata tells us:
            + all the known keys (key id and public key value)
            + which keys belong to each role
            + how many signatures are needed for each role
        - targets metadata tells us:
            + which target files are available (filename, size, hash)
        - snapshots metatadata tells us:
            + which version of the targets-metadata file to expect
        - timestamp metadata tells us:
            + which version of the snapshot-metadata file to expect

        """
        if dir_path is None:
            dir_path = pathlib.Path.cwd() / DEFAULT_META_DIR_NAME
        super().__init__(dir_path=dir_path)
        # top-level roles
        self.root: Optional[Metadata[Root]] = None
        self.targets: Optional[Metadata[Targets]] = None
        self.snapshot: Optional[Metadata[Snapshot]] = None
        self.timestamp: Optional[Metadata[Timestamp]] = None
        # import roles from dir_path, if it exists
        self._import_roles(role_names=TOP_LEVEL_ROLE_NAMES)

    def _import_roles(self, role_names: Iterable[str]):
        """Import roles from metadata files."""
        file_paths = []
        if self.dir_path.exists():
            file_paths = [p for p in self.dir_path.iterdir() if p.is_file()]
        for role_name in role_names:
            # exclude (root) filenames starting with a version
            role_paths = [p for p in file_paths if p.name.startswith(role_name)]
            if role_paths:
                # there should be only one for each role
                setattr(self, role_name, Metadata.from_file(str(role_paths[0])))

    def initialize(
            self, keys: Keys, expiration_days: Optional[RolesDict] = None
    ):
        if expiration_days is None:
            expiration_days = DEFAULT_EXPIRATION_DAYS.copy()
        # based on python-tuf basic_repo.py
        common_kwargs = dict(version=1, spec_version=SPEC_VERSION)
        # role-specific kwargs
        initial_data = {
            Root: dict(
                expires=in_(expiration_days['root']),
                keys=keys.public(),
                roles=keys.roles(),
                # repo is relatively static, no need for consistent snapshots
                consistent_snapshot=False,
            ),
            Targets: dict(
                expires=in_(expiration_days['targets']),
                targets=dict(),
            ),
            Snapshot: dict(
                expires=in_(expiration_days['snapshot']),
                meta={FILENAME_TARGETS: MetaFile(version=1)},
            ),
            Timestamp: dict(
                expires=in_(expiration_days['timestamp']),
                snapshot_meta=MetaFile(version=1),
            ),
        }
        for role_class, role_kwargs in initial_data.items():
            attr_name = role_class.type
            if getattr(self, attr_name) is None:
                # intialize role only if there is no role yet
                setattr(
                    self,
                    attr_name,
                    Metadata(
                        signed=role_class(**common_kwargs, **role_kwargs),
                        signatures=dict(),
                    ),
                )

    def add_or_update_target(
            self,
            local_path: Union[pathlib.Path, str],
            url_path_segments: Optional[List[str]] = None,
    ):
        # based on python-tuf basic_repo.py
        local_path = pathlib.Path(local_path)
        # build url path
        url_path_segments = url_path_segments or []
        url_path_segments.append(local_path.name)
        url_path = '/'.join(url_path_segments)
        target_file_info = TargetFile.from_file(
            target_file_path=url_path, local_path=str(local_path)
        )
        # note we assume self.targets has been initialized
        self.targets.signed.targets[url_path] = target_file_info

    def remove_target(self, local_path: Union[pathlib.Path, str]) -> bool:
        removed = False
        targets_dict = self.targets.signed.targets
        for target_url in targets_dict:
            # assume target filenames only occur once
            if target_url.endswith(local_path.name):
                removed = targets_dict.pop(target_url, None) is not None
                break
        if removed:
            local_path.unlink()
        return removed

    def add_public_key(
            self, role_name: str, public_key_path: Union[pathlib.Path, str]
    ):
        """Import a public key from file and add it to the specified role."""
        # based on python-tuf basic_repo.py
        ssl_key = import_ed25519_publickey_from_file(filepath=str(public_key_path))
        self.root.signed.add_key(
            role=role_name, key=SSlibKey.from_securesystemslib_key(ssl_key)
        )

    def set_signature_threshold(self, role_name: str, threshold: int):
        self.root.signed.roles[role_name].threshold = threshold

    def set_expiration_date(self, role_name: str, days: int):
        role = getattr(self, role_name)
        if hasattr(role, 'signed'):
            role.signed.expires = in_(days)

    def sign_role(
            self, role_name: str, private_key_path: Union[pathlib.Path, str]
    ):
        """
        Sign role using specified private key.

        We sign off on the role.signed part, and the signature is added to
        the role.signatures list.
        """
        # based on python-tuf basic_repo.py
        try:
            # assume unencrypted
            ssl_key = import_ed25519_privatekey_from_file(
                filepath=str(private_key_path), prompt=False
            )
        except CryptoError as e:
            # possibly encrypted: try to import with password
            logger.debug(f'private key import attempt without password failed: {e}')
            ssl_key = import_ed25519_privatekey_from_file(
                filepath=str(private_key_path), prompt=True
            )
        signer = SSlibSigner(ssl_key)
        getattr(self, role_name).sign(signer, append=True)

    def file_path(self, role_name: str, version: Optional[int] = None):
        version_str = ''
        if role_name == Root.type and version is not None:
            # "... all released versions of root metadata files MUST always
            # be provided so that outdated clients can update to the latest
            # available root."
            # https://theupdateframework.github.io/specification/latest/#writing-consistent-snapshots
            version_str = f'{version}.'
        return self.dir_path / self.filename_pattern.format(
            version=version_str, role_name=role_name, suffix=SUFFIX_JSON
        )

    def file_exists(self, role_name: str):
        """
        Return True if any metadata file exists for the specified role,
        ignoring any versions in the filename.
        """
        return any(
            path.exists() for path in self.dir_path.iterdir()
            if path.is_file() and role_name in path.name
        )

    def persist_role(self, role_name: str):
        """
        Save specified role to corresponding metadata file.

        In case of root, make sure "root.json" always represents the latest
        version (in addition to x.root.json).
        """
        # based on python-tuf basic_repo.py (but without consistent snapshots)
        role = getattr(self, role_name)
        file_path = self.file_path(
            role_name=role_name, version=role.signed.version
        )
        role.to_file(
            filename=str(file_path), serializer=JSONSerializer(compact=False)
        )
        if role_name == Root.type:
            # Copy the latest root metadata to 'root.json' (without version),
            # to use as trusted root metadata for distribution with the
            # client. This is convenient, otherwise we would need to modify
            # the version in the filename every time root is updated.
            # Moreover, we can now easily access the latest root metadata,
            # without having to check the version in the filename.
            client_root_file_path = self.file_path(role_name=Root.type)
            client_root_file_path.unlink(missing_ok=True)
            shutil.copy(src=file_path, dst=client_root_file_path)

    def get_latest_archive(self) -> Optional[TargetMeta]:
        """
        Returns TargetMeta for latest archive.

        Note that all pre-release versions are always included: On the repo
        side, there is no difference between final releases an pre-releases.
        Pre-release specifiers are only used on the Client side, to filter
        available updates).
        """
        # Note this is similar to the logic in Client._check_updates, but not
        # exactly the same. Merging the implementations would overcomplicate
        # things.
        latest_archive = None
        # sort by version
        signed_targets = self.targets.signed.targets if self.targets else dict()
        targets = sorted(
            TargetMeta(key) for key in signed_targets.keys()
        )
        # extract only the archives
        archives = [target for target in targets if target.is_archive]
        if archives:
            latest_archive = archives[-1]
        return latest_archive


class Repository(object):
    """High-level tools for repository management."""
    config_filename = '.tufup-repo-config'

    def __init__(
            self,
            app_name: str,
            app_version_attr: Optional[str] = None,
            repo_dir: Union[pathlib.Path, str, None] = None,
            keys_dir: Union[pathlib.Path, str, None] = None,
            key_map: Optional[RolesDict] = None,
            encrypted_keys: Optional[List[str]] = None,
            expiration_days: Optional[RolesDict] = None,
            thresholds: Optional[RolesDict] = None,
    ):
        if repo_dir is None:
            repo_dir = pathlib.Path.cwd() / DEFAULT_REPO_DIR_NAME
        if keys_dir is None:
            keys_dir = pathlib.Path.cwd() / DEFAULT_KEYS_DIR_NAME
        if key_map is None:
            key_map = deepcopy(DEFAULT_KEY_MAP)
        if encrypted_keys is None:
            encrypted_keys = []
        if expiration_days is None:
            expiration_days = DEFAULT_EXPIRATION_DAYS.copy()
        if thresholds is None:
            thresholds = DEFAULT_THRESHOLDS.copy()
        self.app_name = app_name
        self.app_version_attr = app_version_attr
        # force path object and resolve, in case of relative paths
        self.repo_dir = pathlib.Path(repo_dir).resolve()
        self.keys_dir = pathlib.Path(keys_dir).resolve()
        self.key_map = key_map
        self.encrypted_keys = encrypted_keys
        self.expiration_days = expiration_days
        self.thresholds = thresholds
        # keys and roles
        self.keys: Optional[Keys] = None
        self.roles: Optional[Roles] = None
        self.revoked_key_names = []

    @property
    def config_dict(self):
        """dict to be saved to configuration file."""
        # attributes matching __init__ arguments are stored as configuration
        config_items = inspect.signature(self.__init__).parameters.keys()
        return {item: getattr(self, item) for item in config_items}

    @property
    def metadata_dir(self) -> pathlib.Path:
        return self.repo_dir / DEFAULT_META_DIR_NAME

    @property
    def targets_dir(self) -> pathlib.Path:
        return self.repo_dir / DEFAULT_TARGETS_DIR_NAME

    @property
    def app_version(self) -> str:
        # read version from specified module attribute without importing
        version = ''
        if self.app_version_attr:
            version = str(
                setuptools.config.expand.read_attr(self.app_version_attr)  # noqa
            )
        return version

    @classmethod
    def get_config_file_path(cls) -> pathlib.Path:
        return pathlib.Path.cwd() / cls.config_filename

    def save_config(self):
        """Save current configuration."""
        # todo: write directories relative to config file dir?
        file_path = self.get_config_file_path()
        file_path.write_text(
            data=json.dumps(
                self.config_dict, default=str, sort_keys=True, indent=4
            ),
            encoding='utf-8',
        )

    @classmethod
    def load_config(cls) -> dict:
        """Load configuration dict from file."""
        file_path = cls.get_config_file_path()
        config_dict = dict()
        try:
            config_dict = json.loads(file_path.read_text())
        except FileNotFoundError:
            logger.warning(f'config file not found: {file_path}')
        except json.JSONDecodeError:
            logger.warning(f'config file invalid: {file_path}')
        return config_dict

    @classmethod
    def from_config(cls):
        """Create Repository instance from configuration file."""
        instance = cls(**cls.load_config())
        instance._load_keys_and_roles(create_keys=False)
        return instance

    def initialize(self):
        """
        Initialize (or update) the local repository.

        This includes:

        - create directories if they do not exist
        - import public keys from existing files, or create new key pairs
        - import roles from existing metadata files
        - create root metadata file if it does not exist

        Safe to call for existing keys and roles.
        """
        # Ensure dirs exist
        for path in [self.keys_dir, self.metadata_dir, self.targets_dir]:
            path.mkdir(parents=True, exist_ok=True)

        # Load keys and roles
        self._load_keys_and_roles(create_keys=True)

        # Publish root metadata (save 1.root.json and copy to root.json)
        if not self.roles.file_path('root').exists():
            self.publish_changes(private_key_dirs=[self.keys_dir])

    def refresh_expiration_date(self, role_name: str, days: Optional[int] = None):
        if days is None:
            days = self.expiration_days.get(role_name)
        self.roles.set_expiration_date(role_name=role_name, days=days)

    def replace_key(
            self,
            old_key_name: str,
            new_public_key_path: Union[pathlib.Path, str],
            new_private_key_encrypted: bool,
    ):
        """
        Replace an existing key by a new one, e.g. after a key compromise.

        Note the changes are not published yet: call publish_changes() for that.
        """
        self.revoked_key_names = []
        # Load old public key from file to obtain its key_id
        public_key_path = self.keys.public_key_path(key_name=old_key_name)
        old_key_id = import_ed25519_publickey_from_file(
            filepath=str(public_key_path)
        )['keyid']
        # Get new key name from public key path
        new_public_key_path = pathlib.Path(new_public_key_path)  # force path
        # A key may be used for multiple roles, so we check the key id for
        # all roles.
        for role_name in TOP_LEVEL_ROLE_NAMES:
            try:
                # remove old key_id from roles dict, if found, and remove key
                # from keys dict if it is no longer used by any roles
                self.roles.root.signed.revoke_key(
                    role=role_name, keyid=old_key_id
                )
                # move old key name from key_map to revoked_key_map
                self.key_map[role_name].remove(old_key_name)
                # to ensure continuity, changes to root must be signed both
                # with the new key and the old key (in addition to unmodified
                # keys), so we keep track of revoked keys until signed
                self.revoked_key_names.append(old_key_name)
                logger.debug(f'Key revoked for {role_name}: {old_key_name}')
            except ValueError:
                logger.debug(f'{role_name} does not have key {old_key_name}')
            else:
                # add new key_id
                self.add_key(
                    role_name=role_name,
                    public_key_path=new_public_key_path,
                    encrypted=new_private_key_encrypted,
                )

    def add_key(
            self,
            role_name: str,
            public_key_path: Union[pathlib.Path, str],
            encrypted: bool,
    ):
        """
        Register a new public key for the specified role.

        Note the changes are not published yet: call publish_changes() for that.
        """
        public_key_path = pathlib.Path(public_key_path)
        # add new key to metadata
        self.roles.add_public_key(
            role_name=role_name, public_key_path=public_key_path
        )
        # add new key to key map
        key_name = public_key_path.with_suffix('').name
        if key_name not in self.key_map[role_name]:
            self.key_map[role_name].append(key_name)
        # add new key to encrypted keys if necessary
        already_present = key_name in self.encrypted_keys
        if encrypted and not already_present:
            self.encrypted_keys.append(key_name)

    def add_bundle(
            self,
            new_bundle_dir: Union[pathlib.Path, str],
            new_version: Optional[str] = None,
            skip_patch: bool = False,
    ):
        """
        Adds a new application bundle to the local repository.

        An archive file is created from the app bundle, and this file is
        added to the tuf repository. If a previous archive version is found,
        a patch file is also created and added to the repository, unless
        `skip_patch` is True.

        Note the changes are not published yet: call `publish_changes()` for
        that.
        """
        # enforce path object
        new_bundle_dir = pathlib.Path(new_bundle_dir)
        # determine new version
        if new_version is None:
            # todo: should we check for a valid version string?
            new_version = self.app_version
        # create archive from latest app bundle
        logger.info(f'Creating new archive from bundle: {new_bundle_dir}')
        new_archive = make_gztar_archive(
            src_dir=new_bundle_dir,
            dst_dir=self.targets_dir,
            app_name=self.app_name,
            version=new_version,
        )
        logger.info(f'Archive ready: {new_archive}')
        # check latest archive before registering the new one
        latest_archive = self.roles.get_latest_archive()
        if not latest_archive or latest_archive.version < new_archive.version:
            # register new archive
            self.roles.add_or_update_target(local_path=new_archive.path)
            # create patch, if possible, and register that too
            if latest_archive and not skip_patch:
                patch_path = Patcher.create_patch(
                    src_path=self.targets_dir / latest_archive.path,
                    dst_path=self.targets_dir / new_archive.path,
                )
                self.roles.add_or_update_target(local_path=patch_path)

    def remove_latest_bundle(self):
        """
        Removes the *latest* app bundle from the local repository.

        This deletes the bundle's archive file and corresponding patch file
        from the targets directory, and updates the tuf repository metadata
        accordingly.

        Note the changes are not published yet: call publish_changes() for that
        """
        # Get latest archive
        latest_archive = self.roles.get_latest_archive()
        if latest_archive:
            # remove latest archive and corresponding patch
            archive_path = self.targets_dir / latest_archive.target_path_str
            patch_path = archive_path.with_suffix('').with_suffix(SUFFIX_PATCH)
            for target_path in [archive_path, patch_path]:
                removed = self.roles.remove_target(local_path=target_path)
                logger.info(
                    f'target {"removed" if removed else "not found"}: {target_path}'
                )

    def publish_changes(self, private_key_dirs: List[Union[pathlib.Path, str]]):
        """
        Publish all modified roles. That is, if a role has changed w.r.t. to
        the version on disk:

        - update expiration date (if not yet updated)
        - bump version (if not yet bumped)
        - sign
        - save to disk

        If a role has not been modified, it is skipped.
        """
        # todo: implement custom Metadata subclass with extra methods:
        #  modified, set_expiration_date, sign, persist, etc. So we can do
        #  e.g role.set_expiration_date(days=1) instead of passing the
        #  role_name around
        for role_name in ['root', 'targets', 'snapshot', 'timestamp']:
            role = getattr(self.roles, role_name)
            # filename without version is always the latest version
            latest_file_path = self.roles.file_path(
                role_name=role_name, version=None
            )
            # if the file does not exist yet, the role is considered modified,
            # and we don't want to bump version and expiration date again
            modified = True
            expires_bumped = True
            version_bumped = True
            if latest_file_path.exists():
                latest_role = Metadata.from_file(filename=str(latest_file_path))
                modified = role.signed != latest_role.signed
                expires_bumped = role.signed.expires != latest_role.signed.expires
                version_bumped = role.signed.version > latest_role.signed.version
            if modified:
                # set new expiration date
                if not expires_bumped:
                    self.roles.set_expiration_date(
                        role_name=role_name,
                        days=self.expiration_days.get(role_name),
                    )
                # bump version
                if not version_bumped:
                    role.signed.version += 1
                # sign metadata and persist changes
                role.signatures.clear()
                self.threshold_sign(
                    role_name=role_name, private_key_dirs=private_key_dirs
                )
                # update version in dependent metadata
                dependent = None
                if role_name == 'root':
                    # Not all changes to root require a re-sign of the other
                    # metadata files (e.g. we could just add some additional
                    # valid keys). However, to be on the safe side,
                    # we'll force a re-sign cascade by bumping the targets
                    # version. Note this may cause a double version bump for
                    # targets, but that should not matter.
                    if self.roles.file_path(
                            role_name='targets', version=None
                    ).exists():
                        self.roles.targets.signed.version += 1
                elif role_name == 'targets':
                    dependent = self.roles.snapshot.signed.meta[FILENAME_TARGETS]
                elif role_name == 'snapshot':
                    dependent = self.roles.timestamp.signed.snapshot_meta
                if dependent:
                    dependent.version = role.signed.version
                logger.info(f'Published changes for {role_name}.')
            else:
                logger.info(f'No changes detected for {role_name}.')
                # Check if signature count meets threshold
                threshold = self.roles.root.signed.roles[role_name].threshold
                if len(role.signatures) < threshold:
                    logger.info(
                        f'{role_name} threshold not met. Trying to sign...'
                    )
                    signature_count = self.threshold_sign(
                        role_name=role_name, private_key_dirs=private_key_dirs
                    )
                    logger.info(f'Added {signature_count} signatures.')
        # update config if key_map has changed
        if self.config_dict != self.load_config():
            self.save_config()
            logger.info('Config file updated.')

    def threshold_sign(
            self,
            role_name: str,
            private_key_dirs: List[Union[pathlib.Path, str]],
    ) -> int:
        """
        Sign the metadata file for a specific role, and save changes to disk.

        Use this to sign and save without making any changes to the actual
        signed metadata.

        In case of root key rotation, both the old private key and the new
        private key are required.

        Returns the number of signatures created.
        """
        signature_count = 0
        # sign role with all required keys that can be found
        key_names = set(self.key_map.get(role_name, []))
        if role_name == 'root':
            # set ensures uniqueness
            key_names = key_names.union(self.revoked_key_names)
        for key_name in key_names:
            private_key_path = self.keys.find_private_key(
                key_name=key_name, key_dirs=private_key_dirs
            )
            if private_key_path:
                self.roles.sign_role(
                    role_name=role_name,
                    private_key_path=private_key_path,
                )
                signature_count += 1
                if key_name in self.revoked_key_names:
                    self.revoked_key_names.remove(key_name)
                    if key_name in self.encrypted_keys:
                        self.encrypted_keys.remove(key_name)
            else:
                logger.warning(f'Private key not found: {key_name}')
        if not signature_count:
            raise Exception(f'No private keys found for {role_name}.')
        # save changes to disk
        self.roles.persist_role(role_name=role_name)
        return signature_count

    def _load_keys_and_roles(self, create_keys: bool = False):
        # todo: make public, rename load_keys_and_metadata
        if self.keys is None:
            logger.info('Importing public keys...')
            self.keys = Keys(
                dir_path=self.keys_dir,
                encrypted=self.encrypted_keys,
                key_map=self.key_map,
                thresholds=self.thresholds,
            )
            if create_keys:
                # safe to call if keys exist
                self.keys.create()
            logger.info('Public keys imported.')
        if self.roles is None:
            logger.info('Importing metadata...')
            self.roles = Roles(dir_path=self.metadata_dir)
            self.roles.initialize(
                keys=self.keys, expiration_days=self.expiration_days
            )
            logger.info('Metadata imported.')
